/*
 * Copyright 2010, 2011 mapsforge.org
 *
 * This program is free software: you can redistribute it and/or modify it under the
 * terms of the GNU Lesser General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE. See the GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */
package org.mapsforge.android.maps;

import java.io.IOException;
import java.io.InputStream;
import java.net.SocketTimeoutException;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLConnection;
import java.net.UnknownHostException;

import org.apache.http.HttpResponse;
import org.apache.http.client.HttpClient;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.impl.client.DefaultHttpClient;
import org.apache.http.params.HttpConnectionParams;
import org.apache.http.params.HttpParams;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;
import android.graphics.Canvas;
import android.os.Build;

/**
 * A MapGenerator that downloads map tiles from a server. To build an implementation for a certain tile
 * server, extend this class and implement the abstract methods.
 */
abstract class TileDownloadMapGenerator extends MapGenerator {
	private Bitmap decodedBitmap;
	// private int[] pixelColors;
	private Canvas canvas;
	private final StringBuilder stringBuilder;
	private Bitmap tileBitmap;
	private OpenStream openStream;

	/**
	 * Default constructor that must be called by subclasses.
	 */
	TileDownloadMapGenerator() {
		super();
		this.stringBuilder = new StringBuilder(128);

		if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.GINGERBREAD)
			openStream = new OpenStreamGingerbread();
		else
			openStream = new OpenStreamPreGingerbread();

	}

	@Override
	final void cleanup() {
		this.tileBitmap = null;
		if (this.decodedBitmap != null) {
			this.decodedBitmap.recycle();
			this.decodedBitmap = null;
		}
	}

	@Override
	final boolean executeJob(MapGeneratorJob mapGeneratorJob) {
		try {
			getTilePath(mapGeneratorJob.tile, this.stringBuilder);
			// read the data from the tile URL
			InputStream inputStream = openStream.openStream(new URL(this.stringBuilder.toString()));
			if (inputStream == null)
				return false;

			this.decodedBitmap = BitmapFactory.decodeStream(inputStream);
			inputStream.close();

			// check if the input stream could be decoded into a bitmap
			if (this.decodedBitmap == null) {
				return false;
			}

			// copy all pixels from the decoded bitmap to the color array
			// this.decodedBitmap.getPixels(this.pixelColors, 0, Tile.TILE_SIZE, 0, 0,
			// Tile.TILE_SIZE, Tile.TILE_SIZE);
			// this.decodedBitmap.recycle();
			//
			// // copy all pixels from the color array to the tile bitmap
			if (this.tileBitmap != null) {
				canvas.drawBitmap(decodedBitmap, 0, 0, null);
				// this.tileBitmap.setPixels(this.pixelColors, 0, Tile.TILE_SIZE, 0, 0,
				// Tile.TILE_SIZE, Tile.TILE_SIZE);
			}

			this.decodedBitmap.recycle();
			this.decodedBitmap = null;
			return true;
		} catch (UnknownHostException e) {
			Logger.debug(e.getMessage());
			return false;
		} catch (IOException e) {
			Logger.exception(e);
			return false;
		}
	}

	/**
	 * Returns the host name of the tile download server.
	 * 
	 * @return the server name.
	 */
	abstract String getServerHostName();

	/**
	 * Stores the absolute path to the requested tile in the given StringBuilder.
	 * 
	 * @param tile
	 *            the tile for which an image is required.
	 * @param imagePath
	 *            the StringBuilder to store the URL to the image.
	 */
	abstract void getTilePath(Tile tile, StringBuilder imagePath);

	@Override
	final void prepareMapGeneration() {
		this.stringBuilder.setLength(0);
	}

	@Override
	final void setupMapGenerator(Bitmap bitmap) {
		this.tileBitmap = bitmap;
		// this.pixelColors = new int[Tile.TILE_SIZE * Tile.TILE_SIZE];
		this.canvas = new Canvas(bitmap);
	}

	private final static int CONNECTION_TIMEOUT_MS = 1000;
	private final static int CONNECTION_MAX_RETRY_COUNT = 3;

	private abstract class OpenStream {
		public OpenStream() {
		}

		protected abstract void prepareConnection(URL url) throws IOException;

		protected abstract InputStream getInputStream() throws IOException;

		public InputStream openStream(URL url) throws IOException {
			prepareConnection(url);
			InputStream is = null;
			int retryCount = 0;
			while (!isInterrupted() && retryCount < CONNECTION_MAX_RETRY_COUNT) {
				try {
					is = getInputStream();
					break;
				} catch (IOException e) {
					retryCount++;
					if (retryCount < CONNECTION_MAX_RETRY_COUNT)
						continue;
					throw e;
				}
			}

			if (isInterrupted())
				return null;

			if (is == null)
				throw new SocketTimeoutException("Connection timeout.");

			return is;
		}
	}

	private class OpenStreamGingerbread extends OpenStream {
		URLConnection mConnection;

		public OpenStreamGingerbread() {
			super();
		}

		@Override
		protected void prepareConnection(URL url) throws IOException {
			mConnection = url.openConnection();
			mConnection.setConnectTimeout(CONNECTION_TIMEOUT_MS);
			mConnection.setReadTimeout(CONNECTION_TIMEOUT_MS);
		}

		@Override
		protected InputStream getInputStream() throws IOException {
			InputStream is;
			try {
				is = mConnection.getInputStream();
			} catch (IllegalStateException e) {
				throw new IOException(e);
			}

			return is;
		}
	}

	private class OpenStreamPreGingerbread extends OpenStream {
		HttpClient mClient;
		HttpGet mGet;
		HttpResponse mResponse;

		public OpenStreamPreGingerbread() {
			super();
			mClient = new DefaultHttpClient();
			HttpParams params = mClient.getParams();
			HttpConnectionParams.setConnectionTimeout(params, CONNECTION_TIMEOUT_MS);
			HttpConnectionParams.setSoTimeout(params, CONNECTION_TIMEOUT_MS);
		}

		@Override
		protected void prepareConnection(URL url) throws IOException {
			try {
				mGet = new HttpGet(url.toURI());
			} catch (URISyntaxException e) {
				throw new IOException(e.getMessage());
			}
		}

		@Override
		protected InputStream getInputStream() throws IOException {
			mResponse = mClient.execute(mGet);
			return mResponse.getEntity().getContent();
		}
	}
}